import json

from tornado.testing import AsyncTestCase, gen_test
from testnado.service_case_helpers import ServiceCaseHelpers

from tests.helpers.service_helpers import fetch_token
from tornadorax.services.queue_service import QueueService


class TestQueue(ServiceCaseHelpers, AsyncTestCase):

    def setUp(self):
        super(TestQueue, self).setUp()

        def health_handle(handler):
            handler.set_status(204)

        def get_message_handle(handler):
            self.get_message_requests.append(handler.request)
            base_url = handler.request.full_url().split("?")[0]
            next_url = base_url + "?marker=" + str(
                len(self.get_message_requests))
            handler.write({
                "links": [{"rel": "next", "href": next_url}],
                "messages": [{"ttl": 10, "age": 10, "body": {"event": "foo"}}]
            })

        def post_message_handle(handler):
            self.post_message_requests.append(handler.request)
            handler.set_status(201)
            handler.finish({
                "resources": [
                    "/v1/queues/myqueue/messages/1",
                ]
            })

        self.get_message_requests = []
        self.post_message_requests = []

        self.queue_service = self.add_service()
        self.queue_service.add_method("GET", "/v1/health", health_handle)
        self.queue_service.add_method(
            "GET", "/v1/queues/myqueue/messages", get_message_handle)
        self.queue_service.add_method(
            "POST", "/v1/queues/myqueue/messages", post_message_handle)

        self.client = QueueService(
            self.queue_service.url("v1"), fetch_token=fetch_token,
            ioloop=self.io_loop)

    @gen_test
    def test_fetch_messages_returns_empty_list(self):

        def get_message_handle(handler):
            handler.set_status(204)
            handler.finish("")

        self.queue_service.add_method(
            "GET", "/v1/queues/myqueue/messages", get_message_handle)
        self.start_services()

        queue = yield self.client.fetch_queue("myqueue")
        result = yield queue.fetch_messages()
        self.assertEqual("success", result["status"])
        messages = result["messages"]
        self.assertEqual(0, len(messages))

    @gen_test
    def test_fetch_messages_returns_messages_on_success(self):
        self.start_services()
        queue = yield self.client.fetch_queue("myqueue")
        result = yield queue.fetch_messages()
        self.assertEqual("success", result["status"])
        messages = result["messages"]
        self.assertEqual(1, len(messages))
        self.assertEqual({"event": "foo"}, messages[0]["body"])
        self.queue_service.assert_requested(
            "GET", "/v1/queues/myqueue/messages", headers={
                "X-Auth-Token": "TOKEN",
                "Client-Id": queue.receive_client_id})

    @gen_test
    def test_fetch_messages_returns_reason_on_failure(self):
        def fail_get_message(handler):
            handler.set_status(404)
            handler.write("ERROR")

        self.queue_service.add_method(
            "GET", "/v1/queues/myqueue/messages", fail_get_message)
        self.start_services()
        queue = yield self.client.fetch_queue("myqueue")
        result = yield queue.fetch_messages()
        self.assertEqual("error", result["status"])
        self.assertEqual(404, result["code"])
        self.assertEqual("ERROR", result["body"])

    @gen_test
    def test_push_message_succeeds(self):
        self.start_services()

        message = {"event": "foo"}
        queue = yield self.client.fetch_queue("myqueue")
        result = yield queue.push_message(message, ttl=60)
        self.assertEqual("success", result["status"])
        self.assertTrue("resource" in result)

        self.queue_service.assert_requested(
            "POST", "/v1/queues/myqueue/messages", headers={
                "X-Auth-Token": "TOKEN",
                "Client-Id": queue.send_client_id})

        request = self.post_message_requests[0]
        request = json.loads(request.body)
        self.assertEqual([{"ttl": 60, "body": message}], request)

    @gen_test
    def test_push_message_returns_reason_on_failure(self):
        def fail_push_message(handler):
            handler.set_status(404)
            handler.write("ERROR")

        self.queue_service.add_method(
            "POST", "/v1/queues/myqueue/messages", fail_push_message)

        self.start_services()
        queue = yield self.client.fetch_queue("myqueue")
        result = yield queue.push_message("FOOBAR", 60)
        self.assertEqual("error", result["status"])
        self.assertEqual(404, result["code"])
        self.assertEqual("ERROR", result["body"])

    @gen_test
    def test_wait_for_message_returns_after_multiple_fetches(self):
        # placeholder for a little more friendly interface for
        # tornaodo for longer polling

        # message = yield.queue.wait_for_messages()
        pass



LIST_RESPONSE = {
    "loadBalancers": [
        {
            "name": "lb1",
            "id": 1,
            "protocol": "HTTP",
            "port": 80,
            "algorithm": "RANDOM",
            "status": "ACTIVE",
            "nodeCount": 3,
            "virtualIps": [
                {
                    "id": 401,
                    "address": "206.55.130.1",
                    "type": "PUBLIC",
                    "ipVersion": "IPV4"
                }
            ],
            "created": {
                "time": "2010-11-30T03:23:42Z"
            },
            "updated": {
                "time": "2010-11-30T03:23:44Z"
            }
        },
        {
            "name": "lb2",
            "id": 2,
            "protocol": "HTTPS",
            "port": 443,
            "algorithm": "RANDOM",
            "status": "ACTIVE",
            "nodeCount": 4,
            "virtualIps": [
                {
                    "id": 402,
                    "address": "206.55.130.2",
                    "type": "PUBLIC",
                    "ipVersion": "IPV4"
                }
            ],
            "created": {
                "time": "2010-11-30T03:23:42Z"
            },
            "updated": {
                "time": "2010-11-30T03:23:44Z"
            }
        }
    ]
}


CREATE_RESPONSE = {
    "loadBalancer": {
        "name": "a-new-loadbalancer",
        "id": 3,
        "protocol": "HTTP",
        "halfClosed": "true",
        "port": 80,
        "algorithm": "RANDOM",
        "status": "BUILD",
        "timeout": 30,
        "cluster": {
            "name": "cluster1"
        },
        "nodes": [],
        "virtualIps": [
            {
                "address": "206.10.10.210",
                "id": 39,
                "type": "PUBLIC",
                "ipVersion": "IPV4"
            }
        ],
        "connectionLogging": {
            "enabled": False
        }
    }
}

GET_RESPONSE = {
    "loadBalancer": {
        "name": "a-new-loadbalancer",
        "id": 3,
        "protocol": "HTTP",
        "halfClosed": "true",
        "port": 80,
        "algorithm": "RANDOM",
        "status": "BUILD",
        "timeout": 30,
        "cluster": {
            "name": "cluster1"
        },
        "nodes": [],
        "virtualIps": [
            {
                "address": "206.10.10.210",
                "id": 39,
                "type": "PUBLIC",
                "ipVersion": "IPV4"
            }
        ],
        "connectionLogging": {
            "enabled": False
        }
    }
}



import logging
import json
import uuid
import urlparse

from tornado import gen
from tornado.httpclient import AsyncHTTPClient


LOGGER = logging.getLogger("rax:queues")


class QueueService(object):

    def __init__(self, service_url, fetch_token, ioloop):
        self.service_url = service_url
        self.fetch_token = fetch_token
        self.ioloop = ioloop

    @gen.coroutine
    def fetch_queue(self, queue_name):
        # TODO: check it exists, create it, or something. this
        # is pretty simple now.
        queue = Queue(
            self.service_url, queue_name, self.fetch_token, self.ioloop)
        raise gen.Return(queue)


class Queue(object):

    def __init__(self, service_url, queue_name, fetch_token, ioloop):
        parsed_uri = urlparse.urlparse(service_url)
        self.ioloop = ioloop
        self.service_url = service_url
        self.protocol = parsed_uri.scheme
        self.fetch_token = fetch_token
        self.queue = queue_name
        self.client = AsyncHTTPClient(io_loop=self.ioloop)
        self.receive_client_id = uuid.uuid4().hex
        self.send_client_id = uuid.uuid4().hex
        self.next_url = None

    @gen.coroutine
    def fetch_messages(self):
        token = yield self.fetch_token()

        if not self.next_url:
            self.next_url = "{}/queues/{}/messages".format(
                self.service_url, self.queue)

        response = yield self.client.fetch(
            self.next_url, headers={
                "X-Auth-Token": token,
                "Client-Id": self.receive_client_id
            }, raise_error=False)

        if response.code > 399:
            raise gen.Return({
                "status": "error",
                "code": response.code,
                "body": response.body
            })

        if response.code == 204:
            raise gen.Return({
                "status": "success",
                "messages": []
            })

        body = json.loads(response.body)
        next_urls = [l["href"] for l in body["links"] if l["rel"] == "next"]
        self.next_url = urlparse.urljoin(self.service_url, next_urls[0])
        messages = [m for m in body["messages"]]
        raise gen.Return({
            "status": "success",
            "messages": messages
        })

    @gen.coroutine
    def push_message(self, message, ttl):
        token = yield self.fetch_token()

        messages_url = "{}/queues/{}/messages".format(
            self.service_url, self.queue)
        body = json.dumps([{"ttl": ttl, "body": message}])
        response = yield self.client.fetch(
            messages_url, method="POST", body=body, headers={
                "X-Auth-Token": token,
                "Client-Id": self.send_client_id}, raise_error=False)

        if response.code != 201:
            LOGGER.error(
                "Failed to push message: {} {}".format(
                    response.code, response.body))
            raise gen.Return({
                "status": "error",
                "code": response.code,
                "body": response.body
            })

        body = json.loads(response.body)
        raise gen.Return({
            "status": "success",
            "resource": body["resources"][0]
        })

    # TODO: add creation, deletion, broadcast elements, etc.



import json
import hashlib

from tornado import gen
from tornado.concurrent import Future
from tornado.httpclient import AsyncHTTPClient


class StorageService(object):

    def __init__(self, service_url, fetch_token, ioloop):
        self.service_url = service_url
        self.fetch_token = fetch_token
        self.ioloop = ioloop

    @gen.coroutine
    def fetch_container(self, container_name):
        container_url = "{0}/{1}".format(self.service_url, container_name)
        container = StorageContainer(
            container_url, container_name, self.fetch_token,
            ioloop=self.ioloop)
        raise gen.Return(container)


class StorageContainer(object):

    def __init__(self, container_url, name, fetch_token, ioloop):
        self.name = name
        self.container_url = container_url
        self.fetch_token = fetch_token
        self.ioloop = ioloop

    @gen.coroutine
    def fetch_object(self, object_name):
        object_url = "{0}/{1}".format(self.container_url, object_name)
        storage_object = StorageObject(
            object_url, self.name, object_name, self.fetch_token, self.ioloop)
        raise gen.Return(storage_object)


class StorageObject(object):

    def __init__(self, url, container, object_name, fetch_token, ioloop):
        self.object_url = url
        self.container = container
        self.name = object_name
        self.fetch_token = fetch_token
        self.ioloop = ioloop
        self.client = AsyncHTTPClient(io_loop=ioloop)

    @gen.coroutine
    def upload_stream(self, mimetype, writer=None, content_length=0):
        token = yield self.fetch_token()
        writer = writer or BodyWriter
        writer_instance = writer(
            self.object_url, self.container, self.name, mimetype=mimetype,
            token=token, ioloop=self.ioloop, content_length=content_length)
        raise gen.Return(writer_instance)


class BodyWriter(object):

    def __init__(
            self, url, container_name, object_name, token, mimetype,
            ioloop, content_length):
        self.content_length = content_length
        self.transferred_length = 0
        self.ioloop = ioloop
        self.md5sum = hashlib.md5()
        self.headers = {
            "X-Auth-Token": token,
            "Content-type": mimetype
        }

        if content_length > 0:
            self.headers["Content-length"] = str(content_length)

        self.client = AsyncHTTPClient(io_loop=ioloop)
        self.initialized_future = Future()
        self.finish_future = Future()
        self.request_future = self.client.fetch(
            url, method="PUT", body_producer=self, raise_error=False,
            headers=self.headers, request_timeout=0)

    @gen.coroutine
    def __call__(self, write_function):
        self.write_function = write_function
        self.initialized_future.set_result(None)
        yield self.finish_future

    @gen.coroutine
    def write(self, data):
        yield self.initialized_future
        self.md5sum.update(data)
        self.transferred_length += len(data)
        yield self.write_function(data)
        raise gen.Return(len(data))

    @gen.coroutine
    def finish(self):
        self.finish_future.set_result(None)
        response = yield self.request_future

        if response.code not in range(200, 300):
            raise gen.Return({
                "status": "error",
                "code": response.code,
                "body": response.body
            })

        raise gen.Return({
            "status": "success",
            "length": self.transferred_length,
            "md5sum": response.headers.get("ETag")
        })


# 1GB default segment size
DEFAULT_SEGMENT_SIZE = 1024 * 1024 * 1024


class SegmentWriter(object):

    @classmethod
    def with_segment_size(cls, segment_size):
        def chunk_wrapper(*args, **kwargs):
            kwargs.setdefault("segment_size", segment_size)
            return cls(*args, **kwargs)
        return chunk_wrapper

    def __init__(
            self, url, container, object_name, token, mimetype, ioloop,
            content_length, segment_size=DEFAULT_SEGMENT_SIZE):
        self.url = url
        self.segment_size = segment_size
        self.container = container
        self.object_name = object_name
        self.token = token
        self.mimetype = mimetype
        self.content_length = content_length
        self.segment_data = ""
        self.ioloop = ioloop
        self.md5sum = hashlib.md5()
        self.segments = []
        self.segment_indexes = {}
        self.current_segment_number = 0
        self.current_segment = None

    def create_segment(self, segment_name=None):
        if not segment_name:
            self.current_segment_number += 1
            segment_name = "%06d" % (self.current_segment_number)

        segment_name = "segments/{}".format(segment_name)
        segment_url = "{0}/{1}".format(self.url, segment_name)

        segment = BodyWriter(
            segment_url, self.container, self.object_name, self.token,
            "application/video-segment", self.ioloop, content_length=0)

        segment_path = "/{0}/{1}/{2}".format(
            self.container, self.object_name, segment_name)

        self.segment_indexes[id(segment)] = len(self.segments)

        self.segments.append({
            "path": segment_path,
            "etag": "",
            "size_bytes": 0
        })

        return segment

    @gen.coroutine
    def write(self, data):
        if not self.current_segment:
            self.current_segment = self.create_segment()

        self.segment_data += data

        while len(self.segment_data) > self.segment_size:
            extra_bytes = self.segment_data[self.segment_size:]
            self.segment_data = self.segment_data[:self.segment_size]
            yield self.current_segment.write(self.segment_data)
            yield self.close_segment(self.current_segment)
            self.current_segment = self.create_segment()
            self.segment_data = extra_bytes

    @gen.coroutine
    def close_segment(self, segment):
        result = yield segment.finish()
        # TODO: verify and retry
        segment_index = self.segment_indexes[id(segment)]
        self.segments[segment_index]["etag"] = result["md5sum"]
        self.segments[segment_index]["size_bytes"] = result["length"]
        self.md5sum.update(result["md5sum"])
        raise gen.Return(result)

    @gen.coroutine
    def finish(self, dynamic=False):
        if self.current_segment:
            if self.segment_data:
                yield self.current_segment.write(self.segment_data)
            yield self.close_segment(self.current_segment)

        client = AsyncHTTPClient(io_loop=self.ioloop)

        headers = {
            "X-Auth-Token": self.token,
            "Content-type": self.mimetype
        }

        body = json.dumps(self.segments)
        manifest_url = self.url + "?multipart-manifest=put"

        if dynamic:
            # much simpler, but eventually consistent and no explicitness
            headers["X-Object-Manifest"] = "{}/{}/segments".format(
                self.container, self.object_name)
            body = ""
            manifest_url = self.url

        response = yield client.fetch(
            manifest_url, method="PUT", body=body, headers=headers)

        # TODO: verify and retry

        raise gen.Return({
            "etag": response.headers["Etag"],
            "status": "success",
            "md5sum": self.md5sum.hexdigest(),
            "length": sum([s["size_bytes"] for s in self.segments])
        })



from tornadorax.services.queue_service import QueueService
from tornadorax.services.storage_service import StorageService
from tornadorax.services.load_balancer_service import LoadBalancerService


SERVICES = {}


def register_service(service_type, service_class):
    SERVICES[service_type] = service_class


# registering default services included in this project
register_service("rax:queues", QueueService)
register_service("rax:load-balancer", LoadBalancerService)
register_service("object-store", StorageService)


def build_service(
        service_type, service_catalog, fetch_token, ioloop, region=None,
        internal=False):

    if service_type not in SERVICES:
        raise InvalidServiceType(
            "Unregistered service type {0}".format(service_type))

    services = dict([(s["type"], s) for s in service_catalog])

    if service_type not in services:
        raise InvalidServiceType(
            "Server catalog does not contain service type {0}".format(
                service_type))

    # some of this could be left to an internal service classmethod
    # if it turns out these structures deviate widely across providers
    # and services

    endpoints = services[service_type]["endpoints"]
    url_key = "internalURL" if internal else "publicURL"
    regions = dict([(e["region"], e) for e in endpoints])
    region_keys = [e["region"] for e in endpoints]
    region_key = region or region_keys[0]

    if region_key not in regions:
        raise InvalidRegion(
            "Unknown region {0} for service {1}".format(
                region_key, service_type))

    service_url = regions[region_key][url_key]
    return SERVICES[service_type](service_url, fetch_token, ioloop)


class InvalidRegion(Exception):
    pass


class InvalidServiceType(Exception):
    pass



import calendar
import json
import logging
import time

import dateutil.parser
from tornado import gen
from tornado.httpclient import AsyncHTTPClient

from tornadorax.services import service_registry
from tornadorax import utilities


LOGGER = logging.getLogger("rax:identity")


class IdentityClient(object):

    def __init__(self, identity_url, credentials, ioloop):
        self.identity_url = identity_url
        self.credentials = credentials
        self.client = AsyncHTTPClient(io_loop=ioloop)
        self.ioloop = ioloop
        self.service_catalog = None
        self.token = None
        self._token_expires = 0
        self.default_region = None

    @property
    def token_expires(self):
        return self._token_expires

    @token_expires.setter
    def token_expires(self, expiration_string):
        self._token_expires = calendar.timegm(
            dateutil.parser.parse(expiration_string).utctimetuple())

    @gen.coroutine
    def fetch_token(self):
        # FIXME: If the identity service actually goes down for any
        # measurable length of time, these operations and retries may
        # start stacking and hammering the external service far more
        # than desired. Better result would make everyone wait on a
        # single authorization, and everyone gets success / failure

        # giving a minute of buffer for slight clock skew
        if self.token_expires < time.time() + 60:
            self.token = None

        if self.token:
            raise gen.Return(self.token)

        result = yield self.authorize()
        raise gen.Return(result["token"])

    @gen.coroutine
    def authorize(self):
        full_url = self.identity_url + "/v2.0/tokens"
        body = json.dumps({"auth": self.credentials})
        headers = {"Content-type": "application/json"}

        with utilities.gen_retry(self.ioloop) as wait:
            while True:
                response = yield self.client.fetch(
                    full_url, method="POST", body=body, headers=headers,
                    raise_error=False)

                if response.code not in range(500, 600):
                    # an invalid request error or success
                    break

                LOGGER.warning(
                    "Retrying identity request ({0})".format(response.code))

                yield wait()

        if response.code > 299:
            raise gen.Return({
                "status": "error",
                "code": response.code,
                "body": response.body
            })

        auth_response = json.loads(response.body)
        self.token = auth_response["access"]["token"]["id"]
        # this will need to be augmented for other openstack providers
        self.default_region = auth_response["access"].get("user", {}).get(
            "RAX-AUTH:defaultRegion", None)
        self.token_expires = auth_response["access"]["token"]["expires"]
        self.service_catalog = auth_response["access"]["serviceCatalog"]
        raise gen.Return({"status": "success", "token": self.token})

    def build_service(self, service_type, region=None, internal=False):
        if not self.service_catalog:
            raise NoServiceCatalog(
                "No service catalog has been associated with this "
                "identity client. Try yielding authorize() first.")
        region = region or self.default_region
        return service_registry.build_service(
            service_type, self.service_catalog, ioloop=self.ioloop,
            fetch_token=self.fetch_token, region=region, internal=internal)


class NoServiceCatalog(Exception):
    pass



class ServiceError(Exception):
    # raised for 500 or other error code

    def __init__(self, message, code, body):
        self.code = code
        self.body = body
        super(ServiceError, self).__init__(message)


def check_service_response(response, service):
    if response.code >= 400:
        message = "Service {0} failed with {1}\n{2}".format(
            service, response.code, response.body)
        raise ServiceError(message, code=response.code, body=response.body)



import contextlib
import random
import time

from tornado import gen


_MAX_INTERVAL = 10
MAX_BACKOFF_WAIT_TIME = 2 ** _MAX_INTERVAL


def generate_backoff(interval, max_wait=MAX_BACKOFF_WAIT_TIME):
    # slight optimization to make sure we're not going crazy
    if interval > _MAX_INTERVAL:
        interval = _MAX_INTERVAL
    return min((2 ** interval) * random.random(), max_wait)


@contextlib.contextmanager
def gen_retry(ioloop, max_retries=100):

    @gen.coroutine
    def waiter():
        wait_time = time.time() + generate_backoff(waiter.increment)
        yield gen.Task(ioloop.add_timeout, wait_time)
        waiter.increment += 1
        if waiter.increment > max_retries:
            raise MaxRetriesExceeded(
                "Attempted operation {0} times.".format(waiter.increment))

    waiter.increment = 0
    yield waiter


class MaxRetriesExceeded(Exception):
    pass



import json
import hashlib

from tornado.testing import AsyncTestCase, gen_test
from testnado.service_case_helpers import ServiceCaseHelpers

from tests.helpers.service_helpers import fetch_token
from tornadorax.services.storage_service import StorageService
from tornadorax.services.storage_service import SegmentWriter


class TestStorage(ServiceCaseHelpers, AsyncTestCase):

    def setUp(self):
        super(TestStorage, self).setUp()
        self.storage_service = self.add_service()
        self.storage_service.add_method(
            "PUT", "/v1/container/object", object_write_handle)

        self.storage_service.add_method(
            "PUT", "/v1/container/manifest", object_write_handle)

        self.storage_service.add_method(
            "PUT", "/v1/container/manifest/segments/\d+", object_write_handle)

        self.client = StorageService(
            self.storage_service.url("/v1"), fetch_token=fetch_token,
            ioloop=self.io_loop)

    @gen_test
    def test_upload_stream_stores_contents(self):
        self.start_services()

        container = yield self.client.fetch_container("container")
        obj = yield container.fetch_object("object")
        writer = yield obj.upload_stream(mimetype="text/html")
        yield writer.write("CON")
        yield writer.write("TENTS")
        result = yield writer.finish()

        self.assertEqual("success", result["status"])
        self.assertEqual(8, result["length"])
        self.assertEqual(hashlib.md5("CONTENTS").hexdigest(), result["md5sum"])

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/object", headers={
                "X-Auth-Token": "TOKEN", "Content-type": "text/html"})

        self.assertEqual("CONTENTS", request.body)

    @gen_test
    def test_upload_stream_raises_error(self):

        self.storage_service.add_method(
            "PUT", "/v1/container/object", object_write_error_handle)

        self.start_services()
        container = yield self.client.fetch_container("container")
        obj = yield container.fetch_object("object")
        writer = yield obj.upload_stream(mimetype="text/html")
        yield writer.write("CONTENTS")
        result = yield writer.finish()

        self.assertEqual("error", result["status"])
        self.assertEqual(401, result["code"])
        self.assertEqual("ERROR", result["body"])

    @gen_test
    def test_upload_stream_allows_content_length(self):
        self.start_services()
        container = yield self.client.fetch_container("container")
        obj = yield container.fetch_object("object")
        writer = yield obj.upload_stream(
            mimetype="text/html", content_length=8)
        yield writer.write("CONTENTS")
        yield writer.finish()

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/object")
        self.assertEqual("8", request.headers["Content-length"])

    @gen_test
    def test_upload_stream_allows_segmentation(self):
        # big, nasty segment test. should be broken up later, especially
        # with retry, etc.

        self.start_services()
        container = yield self.client.fetch_container("container")
        obj = yield container.fetch_object("manifest")
        segment_writer = SegmentWriter.with_segment_size(4)
        writer = yield obj.upload_stream(
            mimetype="text/html", writer=segment_writer)
        yield writer.write("abe")
        yield writer.write(" lincoln")
        yield writer.write(" wins")
        result = yield writer.finish()
        self.assertEqual("success", result["status"])

        expected = [
            ("abe ", "/container/manifest/segments/000001"),
            ("linc", "/container/manifest/segments/000002"),
            ("oln ", "/container/manifest/segments/000003"),
            ("wins", "/container/manifest/segments/000004")
        ]

        for content, segment_path in expected:
            request = self.storage_service.assert_requested(
                "PUT", "/v1{}".format(segment_path),
                headers={"X-Auth-Token": "TOKEN"})
            self.assertEqual(content, request.body)

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest",
            headers={"X-Auth-Token": "TOKEN"})
        self.assertEqual("put", request.arguments["multipart-manifest"][0])
        self.assertEqual("text/html", request.headers["Content-type"])

        body = json.loads(request.body)
        self.assertEqual(4, len(body))

        for i in range(len(body)):
            segment_info = body[i]
            expected_body, expected_path = expected[i]
            expected_etag = hashlib.md5(expected_body).hexdigest()
            self.assertEqual(expected_path, segment_info["path"])
            self.assertEqual(expected_etag, segment_info["etag"])
            self.assertEqual(4, segment_info["size_bytes"])

    @gen_test
    def test_upload_stream_allows_custom_segments(self):
        self.start_services()
        container = yield self.client.fetch_container("container")
        obj = yield container.fetch_object("manifest")
        # bad segment size so we can ensure it's not using it
        segment_writer = SegmentWriter.with_segment_size(1)
        writer = yield obj.upload_stream(
            mimetype="text/html", writer=segment_writer)

        segment1 = writer.create_segment()
        yield segment1.write("foo")
        yield segment1.write("bar")
        yield segment1.write("one")
        yield writer.close_segment(segment1)
        segment2 = writer.create_segment()
        yield segment2.write("foobar2")
        yield writer.close_segment(segment2)
        result = yield writer.finish()

        self.assertEqual("success", result["status"])

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest/segments/000001")
        self.assertEqual("foobarone", request.body)

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest/segments/000002")
        self.assertEqual("foobar2", request.body)

    @gen_test
    def test_upload_segment_allows_dynamic_segments(self):
        self.start_services()

        container = yield self.client.fetch_container("container")
        obj = yield container.fetch_object("manifest")

        writer = yield obj.upload_stream(
            mimetype="text/html", writer=SegmentWriter)

        segment1 = writer.create_segment("001")
        yield segment1.write("foo")
        yield writer.close_segment(segment1)

        segment2 = writer.create_segment("005")
        yield segment2.write("bar")
        yield writer.close_segment(segment2)

        result = yield writer.finish(dynamic=True)
        self.assertEqual("success", result["status"])

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest/segments/001")
        self.assertEqual("foo", request.body)

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest/segments/005")
        self.assertEqual("bar", request.body)

        request = self.storage_service.assert_requested(
            "PUT", "/v1/container/manifest")
        self.assertEqual(
            "container/manifest/segments",
            request.headers["X-Object-Manifest"])
        self.assertEqual("text/html", request.headers["Content-type"])
        self.assertEqual("", request.body)

    # Need to add tests that verify etags, retry manifests, etc.


def object_write_handle(handler):
    handler.set_status(201)
    handler.set_header(
        "ETag", hashlib.md5(handler.request.body).hexdigest())
    handler.finish()


def object_write_error_handle(handler):
    handler.set_status(401)
    handler.write("ERROR")
